use axum::{
    extract::{FromRequestParts, State},
    http::{header::HeaderMap, request::Parts, StatusCode},
    response::IntoResponse,
    RequestPartsExt,
};
use snafu::prelude::*;
use url::Url;

use crate::{
    error::*,
    state::sessions::*,
    state::{LoggedInUser, RoutableAppState},
};

// TODO: Implement as a typed header
#[derive(Debug)]
pub struct OriginalURI(Url);

#[async_trait]
impl FromRequestParts<RoutableAppState> for OriginalURI {
    type Rejection = Error;

    async fn from_request_parts(
        parts: &mut Parts,
        state: &RoutableAppState,
    ) -> Result<Self, Self::Rejection> {
        let headers: HeaderMap = parts.extract().await.unwrap();
        if let Some(uri) = headers.get("X-Original-URI") {
            trace!("Received original URI: {}", uri.to_str().unwrap());
            // TODO: error
            Ok(OriginalURI(Url::parse(uri.to_str().unwrap()).unwrap()))
        } else {
            // TODO: error
            panic!()
        }
    }
}

impl AsRef<Url> for OriginalURI {
    fn as_ref(&self) -> &Url {
        &self.0
    }
}

#[debug_handler]
pub async fn route(
    user: Option<LoggedInUser>,
    uri: Option<OriginalURI>,
    state: State<RoutableAppState>,
) -> impl IntoResponse {
    debug!("Starting authorization request");
    if uri.is_none() {
        panic!();
    }

    let uri = uri.unwrap();
    let username = user.map(|u| u.username().clone());
    debug!(
        "Requesting authorization for user {:?} URI {}",
        username,
        uri.0.as_str()
    );

    match state.permissions.ssowat_config() {
        Ok(conf) => {
            let mut headers = HeaderMap::new();
            if conf.user_has_permission_for_uri(username.as_ref(), uri.as_ref()) {
                debug!("User {:?} is authorized.", username);
                if let Some(username) = username {
                    headers.insert("Remote-User", username.as_str().parse().unwrap());
                    headers.insert("user", username.as_str().parse().unwrap());
                }
                let res = (StatusCode::OK, headers).into_response();
                debug!("{:?}", res);
                res
            } else {
                debug!("User {:?} is not authorized.", username);
                (StatusCode::FORBIDDEN, headers).into_response()
            }
        }
        Err(e) => {
            panic!()
        }
    }
}
