use axum::{extract::FromRequestParts, http::request::Parts, RequestPartsExt};
use snafu::prelude::*;
use tower_cookies::Cookies;
use yunohost_api::Username;

use crate::{
    error::*,
    state::{sessions::*, RoutableAppState, COOKIE_NAME},
};

#[derive(Clone, Debug)]
pub struct LoggedInUser {
    username: Username,
    timestamp: i64,
}

impl LoggedInUser {
    pub fn new(username: Username, timestamp: i64) -> LoggedInUser {
        LoggedInUser {
            username,
            timestamp,
        }
    }

    pub fn username(&self) -> &Username {
        &self.username
    }

    pub fn timestamp(&self) -> i64 {
        self.timestamp
    }
}

impl From<Session> for LoggedInUser {
    fn from(s: Session) -> LoggedInUser {
        LoggedInUser::new(s.username, s.timestamp)
    }
}

impl std::fmt::Display for LoggedInUser {
    fn fmt(&self, fmt: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.username.fmt(fmt)
    }
}

impl AsRef<Username> for LoggedInUser {
    fn as_ref(&self) -> &Username {
        &self.username()
    }
}

#[async_trait]
impl FromRequestParts<RoutableAppState> for LoggedInUser {
    type Rejection = Error;

    async fn from_request_parts(
        parts: &mut Parts,
        state: &RoutableAppState,
    ) -> Result<Self, Self::Rejection> {
        let cookies: Cookies = parts.extract().await.map_err(|_| Error::Cookie)?;
        trace!(
            "[SESSION:{}/login] Checking if the user has a valid session",
            COOKIE_NAME
        );
        let session_cookie = if let Some(session_cookie) = cookies.get(COOKIE_NAME) {
            trace!(
                "[SESSION:{}/login] User claims to have a session with cookie: {}",
                COOKIE_NAME,
                &session_cookie
            );
            session_cookie
        } else {
            trace!("[SESSION:{}/login] No current session cookie", COOKIE_NAME);
            return Err(Error::Session {
                source: SessionError::NoSession {
                    cookie_name: COOKIE_NAME.to_string(),
                },
            });
        };

        if let Some(session) = state
            .sessions
            .verify_cookie_session(session_cookie.value())
            .await
            .context(SessionSnafu)?
        {
            debug!(
                "[SESSION:{}/login] User {} resumed session.",
                COOKIE_NAME, &session.username
            );
            return Ok(session.into());
        } else {
            trace!(
                "[SESSION:{}/login] User session is invalid or has expired.",
                COOKIE_NAME
            );
            return Err(Error::Session {
                source: SessionError::InvalidOrExpired {
                    cookie_name: COOKIE_NAME.to_string(),
                },
            });
        }
    }
}
