use ring::{hmac, rand};
use snafu::prelude::*;
use std::borrow::Cow;
use tokio::sync::RwLock;
use tower_cookies::Cookie;
use yunohost_api::Username;

use crate::utils::time::now;

/// An error related to session management
#[derive(Debug, Snafu)]
#[snafu(visibility(pub))]
pub enum SessionError {
    #[snafu(display("Some cryptographic operation failed due to evil gnomes"))]
    Crypto,
    #[snafu(display("Malformed cookie: {}", content))]
    MalformedCookie { content: String },
    #[snafu(display("Malformed hex cookie signature: {}", sig))]
    MalformedCookieSig {
        sig: String,
        source: hex::FromHexError,
    },
    #[snafu(display("Malformed cookie UNIX timestamp: {}", timestamp))]
    MalformedCookieTimestamp {
        timestamp: String,
        source: std::num::ParseIntError,
    },
    #[snafu(display("Malformed cookie username (empty)"))]
    MalformedCookieUsername { source: yunohost_api::Error },
    #[snafu(display("Invalid or expired session {}", cookie_name))]
    InvalidOrExpired { cookie_name: String },
    #[snafu(display("No session {}", cookie_name))]
    NoSession { cookie_name: String },
}

/// Holds the currently active cookie-based user sessions for a certain cookie type.
/// Sessions are automatically invalidated when the application is restarted,
/// as the `secret` material is regenerated randomly. They are also invalidated
/// after some time.
pub struct SessionManager {
    /// The time of start, used for invalidated expired sessions
    pub start_time: i64,
    /// The secret material used for signing/encrypting cookies
    pub secret: hmac::Key,
    /// The list of currently valid cookies. Some of them may have been invalidated,
    /// but they are stored in a different list to guarantee fast race-free access to this one.
    pub cookies: RwLock<Vec<Session>>,
    /// The list of invalidated cookies, due to logout or other purging mechanisms.
    /// This list is behind a mutex (RwLock) to prevent race conditions
    pub invalidated_cookies: RwLock<Vec<Session>>,
    /// Expiration duration for set cookies, in seconds
    pub expiration_secs: u64,
}

impl SessionManager {
    pub fn new() -> Result<SessionManager, SessionError> {
        //let rng = Rng::new();
        let rng = rand::SystemRandom::new();
        let key = hmac::Key::generate(hmac::HMAC_SHA256, &rng).map_err(|_| SessionError::Crypto)?;

        Ok(SessionManager {
            start_time: now(),
            secret: key,
            cookies: RwLock::new(Vec::new()),
            invalidated_cookies: RwLock::new(Vec::new()),
            // TODO: make expiration configurable
            expiration_secs: 3600 * 24 * 7,
        })
    }

    /// Verifies that a submitted cookie is a valid session. Returns:
    /// - Err(_) if the cookie format is really wrong
    /// - Ok(Some(Session)) if the user is still logged in
    /// - Ok(None) if the user is no longer lgoged in (invalid/expired cookie)
    pub async fn verify_cookie_session(
        &self,
        cookie_claim: &str,
    ) -> Result<Option<Session>, SessionError> {
        // First check the expiration of the claimed cookie.
        // If the timestamp was messed with, it will fail further verification.
        if let Some(claimed_timestamp) = Session::has_expired(cookie_claim, self.expiration_secs)? {
            // The claimed timestamp is still valid. Check if we ever had that cookie in memory.
            // If the server was restarted, user will have to login again.
            if let Some(valid_cookie) = self
                .find_cookie(&self.cookies, claimed_timestamp, &cookie_claim)
                .await
            {
                // Make sure the session hasn't been invalidated
                if let Some(_invalidated_cookie) = self
                    .find_cookie(&self.invalidated_cookies, claimed_timestamp, &cookie_claim)
                    .await
                {
                    // User has logged out or been removed from the system
                    Ok(None)
                } else {
                    // User is still logged in!
                    Ok(Some(valid_cookie))
                }
            } else {
                // User doesn't have an active session
                Ok(None)
            }
        } else {
            // Claimed Cookie timestamp has expired
            return Ok(None);
        }
    }

    /// Generates a new valid cookie inside the `SessionManager`, and returns:
    /// - the cookie name to be set
    /// - the cookie content that can be sent to a client
    pub async fn make_session(&self, cookie_name: &str, username: &Username) -> Session {
        let now = now();
        let signable_payload = format!("{now}:{cookie_name}:{username}");
        let signed_payload = hmac::sign(&self.secret, signable_payload.as_bytes())
            .as_ref()
            .to_vec();
        let cookie_payload = format!("{}:{}", signable_payload, hex::encode(&signed_payload),);

        let cookie = Session {
            cookie_name: cookie_name.to_string(),
            timestamp: now,
            username: username.clone(),
            signature: signed_payload.clone(),
            content: cookie_payload.clone(),
        };

        {
            // We don't want to block too long the cookie jar
            // So it's only locked for this block (then dropped)
            let mut jar = self.cookies.write().await;
            jar.push(cookie.clone());
        }

        cookie
    }

    /// Invalidated a previously-set session cookie
    pub async fn invalidate_session(&self, timestamp: i64, content: &str) -> bool {
        if let Some(cookie) = self.find_cookie(&self.cookies, timestamp, content).await {
            debug!(
                "[SESSION:{}] User {} is logging out",
                &cookie.cookie_name, &cookie.username
            );
            // Make sure the cookie was not already invalidated to avoid DOSing our invalidated cookie storage
            if self
                .find_cookie(&self.invalidated_cookies, timestamp, content)
                .await
                .is_none()
            {
                let mut jar = self.invalidated_cookies.write().await;
                jar.push(cookie);
                return true;
            }
        }

        // User was not logged in, or session was already invalidated
        false
    }

    /// Helper method to find a cookie with a specific timestamp, name and username in a cookie jar
    async fn find_cookie(
        &self,
        jar: &RwLock<Vec<Session>>,
        timestamp: i64,
        content: &str,
    ) -> Option<Session> {
        jar.read()
            .await
            .iter()
            .find(|cookie| {
                // First compare the timestamp (cheapest operation to invalidate the match)
                cookie.timestamp == timestamp && cookie.content == content
            })
            .cloned()
    }
}

/// A signed/encrypted as stored in memory.
#[derive(Clone, Debug)]
pub struct Session {
    /// The POSIX timetamp this cookie was created at
    pub timestamp: i64,
    /// The cookie name
    pub cookie_name: String,
    /// The username for which this cookie is valid
    pub username: Username,
    /// The cryptographic signature going with the cookie
    pub signature: Vec<u8>,
    /// The stringy representation of the cookie, to be received/sent with clients
    pub content: String,
}

impl Session {
    /// Extrats the timestamp from a stringy cookie
    pub fn timestamp(cookie: &str) -> Result<i64, SessionError> {
        let (timestamp, _rest) = cookie.split_once(':').context(MalformedCookieSnafu {
            content: cookie.to_string(),
        })?;
        let timestamp: i64 = timestamp
            .parse()
            .context(MalformedCookieTimestampSnafu { timestamp })?;

        Ok(timestamp)
    }

    /// Verifies whether a given cookie string has expired, before parsing it entirely. Returns:
    /// - Err(SessionError) when the cookie is malformed
    /// - Ok(Some(timestamp)) when the cookie is still valid
    /// - Ok(None) when the cookie has expired
    /// Will error if the cookie is misformed.
    pub fn has_expired(cookie: &str, expiration_secs: u64) -> Result<Option<i64>, SessionError> {
        let timestamp = Self::timestamp(cookie)?;

        if let Some(expiration) = timestamp.checked_add_unsigned(expiration_secs) {
            if expiration >= now() {
                Ok(Some(timestamp))
            } else {
                Ok(None)
            }
        } else {
            // Addition overflowed. System clock is broken, expiration is set too high, or the client tried to trick us.
            Ok(None)
        }
    }

    /// The typed [`yunohost_api::Username`] for which this cookie is deemed valid.
    pub fn username(&self) -> &Username {
        &self.username
    }

    pub fn cookie(&self) -> Cookie<'static> {
        Cookie::build(
            Cow::Owned(self.cookie_name.to_string()),
            Cow::Owned(self.content.to_string()),
        )
        .path(Cow::Owned("/".to_string()))
        .finish()
    }
}
