use file_owner::PathExt;
use snafu::prelude::*;
use tokio::{fs::set_permissions, task::spawn_blocking};

use std::{fs::Permissions, os::unix::fs::PermissionsExt, path::Path};

use crate::error::*;

pub struct FSPermissions {
    pub owner: Option<String>,
    pub group: Option<String>,
    pub mode: Option<u32>,
}

impl FSPermissions {
    pub fn new() -> FSPermissions {
        FSPermissions {
            owner: None,
            group: None,
            mode: None,
        }
    }

    pub fn chown(mut self, owner: &str) -> Self {
        self.owner = Some(owner.to_string());
        self
    }

    pub fn chgrp(mut self, group: &str) -> Self {
        self.group = Some(group.to_string());
        self
    }

    pub fn chmod(mut self, mode: u32) -> Self {
        self.mode = Some(mode);
        self
    }

    pub async fn apply_to(&self, path: &Path) -> Result<(), Error> {
        if let Some(mode) = self.mode {
            set_permissions(path, Permissions::from_mode(mode))
                .await
                .context(PermissionsSnafu {
                    path: path.to_path_buf(),
                })?;
        }

        if let Some(owner) = &self.owner {
            let owner = owner.to_string();
            let path = path.to_path_buf();
            let _ = spawn_blocking(move || -> Result<(), Error> {
                Ok(path
                    .set_owner(owner.as_str())
                    .context(PermissionsChownSnafu {
                        path: path.to_path_buf(),
                    })?)
            })
            .await
            .context(TokioTaskSnafu)?;
        }

        if let Some(group) = &self.group {
            let group = group.to_string();
            let path = path.to_path_buf();
            let _ = spawn_blocking(move || -> Result<(), Error> {
                Ok(path
                    .set_group(group.as_str())
                    .context(PermissionsChgrpSnafu {
                        path: path.to_path_buf(),
                    })?)
            })
            .await
            .context(TokioTaskSnafu)?;
        }

        Ok(())
    }
}
