use serde::Deserialize;
use snafu::prelude::*;

use std::os::unix::fs::MetadataExt;
use std::path::{Path, PathBuf};
use std::sync::RwLock;

use crate::error::*;

// TODO: Can we return reference to the inner data without cloning?
// Cloning is MUCH cheaper than reading from disk so it's not really a problem
// TODO: Should we TryFrom<Vec<u8>> so we support more than JSON/FromStr?
pub trait Cachable: Clone + std::fmt::Debug + for <'a> Deserialize<'a> + Default {}

impl<T: Clone + std::fmt::Debug + for<'a> Deserialize<'a> + Default> Cachable for T {}

#[derive(Debug)]
pub struct JsonCacheInner<T: Cachable> {
    mtime: i64,
    content: T,
}

impl<T: Cachable> JsonCacheInner<T> {
    pub fn new() -> JsonCacheInner<T> {
        JsonCacheInner {
            mtime: i64::MIN,
            content: T::default(),
        }
    }

    pub fn get(&self) -> T {
        self.content.clone()
    }
}

#[derive(Debug)]
pub struct JsonCache<T: Cachable> {
    path: PathBuf,
    inner: RwLock<JsonCacheInner<T>>,
}

impl<T: Cachable> JsonCache<T> {
    /// Prepares a cached JSON file. Does not load from disk.
    pub fn new(path: &Path) -> JsonCache<T> {
        JsonCache {
            path: path.to_path_buf(),
            inner: RwLock::new(JsonCacheInner::new()),
        }
    }

    /// Loads a cached JSON file from disk to memory
    pub fn load(path: &Path) -> Result<JsonCache<T>, Error> {
        let cache = Self::new(path);
        cache.reload(0)?;
        Ok(cache)
    }

    pub fn get(&self) -> Result<T, Error> {
        if let Some(new_mtime) = self.stale()? {
            self.reload(new_mtime)?;
        }

        let lock = self.inner.read().unwrap();
        let content = lock.get();
        
        Ok(content)
    }

    pub fn stale(&self) -> Result<Option<i64>, Error> {
        let metadata = self.path.metadata().context(ReadFileSnafu { path: self.path.clone() })?;
        let inner = self.inner.read().unwrap();
        let mtime = metadata.mtime();
        if mtime != inner.mtime {
            // Needs reloading
            Ok(Some(mtime))
        } else {
            // Data still valid
            Ok(None)
        }
    }

    fn reload(&self, mtime: i64) -> Result<(), Error> {
        let mut inner = self.inner.write().unwrap();
        let content = std::fs::read_to_string(&self.path).context(ReadFileSnafu { path: self.path.clone() })?;
        //let content = T::from_str(&content)?;
        let content: T = serde_json::from_str(&content).context(InvalidJsonSnafu { path: self.path.clone() })?;
        inner.content = content;
        inner.mtime = mtime;
        Ok(())
    }
}
